/**********************************************************
 * Version $Id$
 *********************************************************/
#ifndef __AUSWERT_FOREACH_H 
#define __AUSWERT_FOREACH_H 


#include "auswert_zuweisung.h"



extern void ausfuehren_anweisung(T_AnweisungList& a);

void ausfuehren_foreach(BBForEach& f);




#endif
