/* config.h.  Generated from config.h.in by configure.  */
#if defined(_WIN32)
#define isnan _isnan
#define copysign _copysign
#define rint (int)
#define M_PI 3.14159265358979323846
#define NO_TIMER
#endif
#define TRILIBRARY
